/***************************************************************************
 *
 * Copyright (c) 2002 Promise Technology, Inc.  All rights reserved.
 * No part of this document may be reproduced or transmitted in any form or
 * by any means, electronic or mechanical, for any purpose, without the
 * express written permission of Promise Technology, Inc.
 *
 *
 * cam_isr.c - contains declaration for global variables
 *
 ***************************************************************************/
#include "cam_con.h"
#include "cam_def.h"
#include "cam_imp.h"
#include "cam_macr.h"
#include "cam_gb.h"
#include "cam_exp.h"

void CAM_NCQ_ISR(U8 Adapter_ID,U8 SEQID)
{
	PSEQ_INFO		pCurSEQ_INFO;
	U8 				ModuleType;
	U8				i;
	PCAM_PACKET		ultmpATAPacketAddress;
	PATA_TASK 		pATATask;
	U8 				SEQerrorFlag=FALSE;
	U8 				SEQCntCtrlErr;
	U8 				bChannelNumber;
	U32 			tmpData;
	U32 			SActiveData;
	U8 				TAG;
	U8			CRC_err=FALSE;

	pCurSEQ_INFO = &(gSEQ_INFO[Adapter_ID][SEQID]);
	ModuleType= (U8)pCurSEQ_INFO->Owner;
	bChannelNumber = Adapter_ID*MAX_ATA_CHNL+ModuleType;
	
	SEQerrorFlag = FALSE;
						
	/* check error from Sequence Counter Control Register */
	ReadPortUchar(HostBaseAddr[Adapter_ID], (SEQID*4)+2, SEQCntCtrlErr);
	SEQCntCtrlErr &= (1<<ModuleType);
				
	if (SEQCntCtrlErr) {
		SEQerrorFlag = TRUE;
	}
	else { /* non error */
		if(gAdapterConfig[Adapter_ID].TapeOutVersion < REVISION_2){/* ASIC tapout prior revision 2 */
			if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) {/* SATA */
				#ifdef	_MMIO_
					ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDESATAError, tmpData );
				#else
					ReadIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDESATAError, tmpData );
				#endif
				if (tmpData & CRC_ERR) {	/* CRC error */
					SEQerrorFlag = TRUE;
					CRC_err=TRUE;
				}
			}
		}
	}
												
	/* if no error(SATA error and packet error) and no data or ide register status need to transfer(PIO, ATA_STATUS_RETURN or ATAPI ), try to execute other packets.*/
	if (!(SEQerrorFlag == TRUE)) {
		if (pCurSEQ_INFO->GO==TRUE) {
			if (ATAModuleNeedPush[bChannelNumber]==TRUE){
				/* start ata module */
				StartATAOperation(bChannelNumber);
			}
		}
	}
	
	if (CRC_err==TRUE)
		SActiveData = 0;	/* no commands complete */
	else {
		/* get finish TAGs */
		ReadIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_SActive, SActiveData);
		SActiveData = (gChnlDrvControlConfig[bChannelNumber].UsedTAG & (~SActiveData));
	}
		
	/* when NCQ running, can't write FPDMA Control/Status Register whole DWORD */
	/* only clear the FPDMA Control/Status Register bit 10, 11*/
	ReadIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_FPDMA_Ctrl_Status, tmpData);
	tmpData |= FPDMA_Ctrl_Status_DMAsetup_INT_flag | FPDMA_Ctrl_Status_SetDb_INT_flag;
	WriteIndexUchar(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_FPDMA_Ctrl_Status+1, (U8) (tmpData>> 8));

	if (SEQerrorFlag == TRUE) { /* error */
	
		camStallExecution(1);
		
		/* get ATA Global control & Status Register */
		ReadPortUlong(ATABaseAddr[bChannelNumber], ATA_STATUS_REGISTER_OFFSET, tmpData);
			
		/* soft reset ATA Module after getting status. */
		ATASoftResetModule(bChannelNumber);
		NeedCheckFIFO[bChannelNumber] = TRUE;
	}
				
	for (i=0;i<MAX_TAG;i++) {
		if (!((SActiveData >> i) & 0x01))
			continue;
		TAG=i;
		gChnlDrvControlConfig[bChannelNumber].UsedTAG &= ~(1 << TAG);
		gChnlDrvControlConfig[bChannelNumber].FreeTAG |= (1 << TAG);
		/* get PCAM_ATA_PACKET from SEQ_INFO */
		ultmpATAPacketAddress=GetATA_PacketFromSEQ_INFO(bChannelNumber,SEQID,TAG);
		pATATask=(PATA_TASK)ultmpATAPacketAddress->PacketHead.PTA;
		pATATask->ATAStatus= 0;		/* default no error */	

		/* free sg resource */
		CAMFreeATASG(bChannelNumber, (PVOID)ultmpATAPacketAddress->PacketHead.PSG);	
		/* free packet resource. */
		CAMFreeCamPacket(bChannelNumber, (PVOID)ultmpATAPacketAddress);				
		pATATask->callback(pATATask);	
	}
	
	if (SEQerrorFlag == TRUE) { /* error */
	
		PCAM_PACKET ultmpATAPacketAddressHead;

		if ((tmpData>>8) & ATAERR_DRV) {/* device error */
			TAG=NCQ_Read_Log_Ext_Page10(bChannelNumber);
		}
		else
			TAG=U8NULL;
		
		/* if can't get the error TAG, hard reset and abort all tasks */
		if (TAG==U8NULL)
		{
			U8	receiveStatus;
			
			/* disable receive task.*/
			receiveStatus=ATAModuleReceiveTask[bChannelNumber];
			ATAModuleReceiveTask[bChannelNumber]=FALSE;
				
			ultmpATAPacketAddressHead = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;
			pCurSEQ_INFO->SEQID2PacketHead=(PVOID)PTR_NULL;
			pCurSEQ_INFO->SEQID2PacketTail=(PVOID)PTR_NULL;			
			/* return all execute commands with error */
			Return_Execute_ATAPackets(bChannelNumber,(PCAM_PACKET)ultmpATAPacketAddressHead,(U16)((tmpData >> 8)|ATA_ERROR) );
			ultmpATAPacketAddressHead=(PCAM_PACKET)PTR_NULL;
			ATAModuleBusy[bChannelNumber]=FALSE;	
			gChnlDrvControlConfig[bChannelNumber].FreeTAG = 0xFFFFFFFF;
			gChnlDrvControlConfig[bChannelNumber].UsedTAG = 0x0;
			
			/* do hard reset so ASIC can restart again */
			/*CAM_Reset_ReinitATA(bChannelNumber);*/
			ATASoftResetModule(bChannelNumber);
			ATAModuleReceiveTask[bChannelNumber]=receiveStatus;
						
		}
		else
		{
			gChnlDrvControlConfig[bChannelNumber].UsedTAG &= ~(1 << TAG);
			gChnlDrvControlConfig[bChannelNumber].FreeTAG |= (1 << TAG);
			
			/* get PCAM_ATA_PACKET from SEQ_INFO */
			ultmpATAPacketAddress=GetATA_PacketFromSEQ_INFO(bChannelNumber,SEQID,TAG);
			if (ultmpATAPacketAddress != (PCAM_PACKET)PTR_NULL) {
				
				ultmpATAPacketAddressHead = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;
				pCurSEQ_INFO->SEQID2PacketHead=(PVOID)PTR_NULL;
				pCurSEQ_INFO->SEQID2PacketTail=(PVOID)PTR_NULL;
				/* back undo execute packets to wait queue. */
				Back_All_Undone_ATAPackets(bChannelNumber,(PCAM_PACKET)ultmpATAPacketAddressHead);
				ATAModuleBusy[bChannelNumber]=FALSE;
				gChnlDrvControlConfig[bChannelNumber].FreeTAG = 0xFFFFFFFF;
				gChnlDrvControlConfig[bChannelNumber].UsedTAG = 0x0;
				ultmpATAPacketAddressHead=(PCAM_PACKET)PTR_NULL;
								
				pATATask=(PATA_TASK)ultmpATAPacketAddress->PacketHead.PTA;
	
				pATATask->ATAStatus = (U16)(tmpData >> 8);
				pATATask->ATAStatus &= ATAERR_HWERROR;
				pATATask->ATAStatus |= ATA_ERROR;						
	
				GetStatus(ATABaseAddr[pATATask->bCh2Chip], pATATask->Cmd.ATACmd.bStatus);
				
				if (pATATask->Cmd.ATACmd.bStatus & IDE_STATUS_BUSY){
					pATATask->ATAStatus &= ~ATAERR_DRV;
					pATATask->ATAStatus |= ATA_TIMEOUT;
				}
				else{				
					/* set disk status */
					if (pATATask->ATACmdFlag & ATA_DEVICE)									
						ATAErrorStatus(pATATask);
				}
				/* free sg resource */
				CAMFreeATASG(bChannelNumber, (PVOID)ultmpATAPacketAddress->PacketHead.PSG);	
				/* free packet resource. */
				CAMFreeCamPacket(bChannelNumber, (PVOID)ultmpATAPacketAddress);
				pATATask->callback(pATATask);	
			}
			else {
				ultmpATAPacketAddressHead = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;
				pCurSEQ_INFO->SEQID2PacketHead=(PVOID)PTR_NULL;
				pCurSEQ_INFO->SEQID2PacketTail=(PVOID)PTR_NULL;				
				/* return all execute commands with error */
				Return_Execute_ATAPackets(bChannelNumber,(PCAM_PACKET)ultmpATAPacketAddressHead,(U16)((tmpData >> 8)|ATA_ERROR));
				ATAModuleBusy[bChannelNumber]=FALSE;
				ultmpATAPacketAddressHead=(PCAM_PACKET)PTR_NULL;	
				gChnlDrvControlConfig[bChannelNumber].FreeTAG = 0xFFFFFFFF;
				gChnlDrvControlConfig[bChannelNumber].UsedTAG = 0x0;								
			}
		}	

	}

	/* must check module is need to push, By it maybe pushed when no SQEID. */
	for (i=(Adapter_ID*MAX_ATA_CHNL);i<((Adapter_ID+1)*MAX_ATA_CHNL);i++) {
		if (ATAModuleNeedPush[i]==TRUE){
			/* start ata module */
			StartATAOperation(i);
				break;
		}
	}
	return;
}

U8 CAM_wait_until_DMAMachine_Idle(U8 bChannelNumber)
{
	U32 tmpData;
	U16 i,j;

	for (i=0;i<100;i++) {
		for (j=0;j<100;j++) {
			/* check DAMMachine is Idle(0 means idel). */
			WriteIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_Internel_debug_1, 0x00);
			ReadIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_Internel_debug_2, tmpData);
			if (!tmpData)
				return TRUE;
		}
		camUStallExecution(1);
	}

	return FALSE;
}


void CAM_QUEUED_ISR(U8 Adapter_ID,U8 SEQID)
{
	PSEQ_INFO		pCurSEQ_INFO;
	U8 				ModuleType;
	U8				i;
	PCAM_PACKET		ultmpATAPacketAddress;
	PATA_TASK 		pATATask;
	U8 				SEQCntCtrlErr;
	U8 				bChannelNumber;
	U8 				TAG;
	U8				Status;
	U32 			tmpData;
	U8				CRC_err=FALSE, ServiceFlag;
	
	U32				StatusPollingCount;
	U8				ErrorOccurFlag;			

	pCurSEQ_INFO = &(gSEQ_INFO[Adapter_ID][SEQID]);
	ModuleType= (U8)pCurSEQ_INFO->Owner;
	bChannelNumber = Adapter_ID*MAX_ATA_CHNL+ModuleType;
	
	/* check error from Sequence Counter Control Register */
	ReadPortUchar(HostBaseAddr[Adapter_ID], (SEQID*4)+2, SEQCntCtrlErr);
	SEQCntCtrlErr &= (1<<ModuleType);

	if (!SEQCntCtrlErr) { /* non error */
		if(gAdapterConfig[Adapter_ID].TapeOutVersion < REVISION_2){/* ASIC tapout prior revision 2 */
			if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) {/* SATA */
				#ifdef	_MMIO_
					ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDESATAError, tmpData );
				#else
					ReadIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDESATAError, tmpData );
				#endif
				if (tmpData & CRC_ERR) {	/* CRC error */
					CRC_err=TRUE;
					SEQCntCtrlErr=TRUE;
				}
			}
		}
	}
					
	if (!SEQCntCtrlErr) {

		/* read status */
		Status = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEStatus);
		if (Status & IDE_STATUS_BUSY)
			return;

		if (gChnlDrvControlConfig[bChannelNumber].TCQTransferData==TRUE) {				
			/* command complete */
			/* get the TAG */
			if(gChnlDrvControlConfig[bChannelNumber].CurrentXferTAG == U8NULL){
				TAG = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDESectorCount);
				TAG = (TAG >> 3);
			}
			else
				TAG = gChnlDrvControlConfig[bChannelNumber].CurrentXferTAG;
				
			gChnlDrvControlConfig[bChannelNumber].UsedTAG &= ~(1 << TAG);
			gChnlDrvControlConfig[bChannelNumber].FreeTAG |= (1 << TAG);
			gChnlDrvControlConfig[bChannelNumber].CurrentXferTAG = U8NULL;
			
			/* get PCAM_ATA_PACKET from SEQ_INFO */
			ultmpATAPacketAddress=GetATA_PacketFromSEQ_INFO(bChannelNumber,SEQID,TAG);
			if (ultmpATAPacketAddress == (PCAM_PACKET)PTR_NULL)
				return;
							
			gChnlDrvControlConfig[bChannelNumber].TCQTransferData = FALSE;
			pATATask=(PATA_TASK)ultmpATAPacketAddress->PacketHead.PTA;
			pATATask->ATAStatus= 0;		/* default no error */	

			CAM_wait_until_DMAMachine_Idle(bChannelNumber);
			
			/* free sg resource */
			CAMFreeATASG(bChannelNumber, (PVOID)ultmpATAPacketAddress->PacketHead.PSG);	
			/* free packet resource. */
			CAMFreeCamPacket(bChannelNumber, (PVOID)ultmpATAPacketAddress);
			pATATask->callback(pATATask);	
		}

		if (gChnlDrvControlConfig[bChannelNumber].TCQTransferData != TRUE) {	
			ServiceFlag = FALSE;
			StatusPollingCount = 0;
			ErrorOccurFlag = FALSE;
			if (Status & IDE_STATUS_SERVICE) { /* Service release case */
				ServiceFlag = TRUE;
			}
			else{
				if(gChnlDrvControlConfig[bChannelNumber].IsWDHardDrive != TRUE){/* not Hitachi */
					if((ATAModuleBusy[bChannelNumber] == TRUE) && (ATAModuleNeedPush[bChannelNumber] != TRUE)){
						while(1) {
							if(StatusPollingCount >= 100000){	/* polling above 1 sec */
								ErrorOccurFlag = TRUE;
								break;
							}

							if (!(Status & IDE_STATUS_BUSY)) {
								if (Status & IDE_STATUS_SERVICE) {
									ServiceFlag = TRUE;
									break;
								}
							}
							camUStallExecution(1);
							Status = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEStatus);
							StatusPollingCount++;
						}		
					}
				}
			}
	
			if(ServiceFlag == TRUE){
				gChnlDrvControlConfig[bChannelNumber].TCQTransferData=TRUE;
				/* send service command */
				IssueService(bChannelNumber,0,IDE_COMMAND_SERVICE);
	
				Status = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEStatus);
				StatusPollingCount = 0;
				ErrorOccurFlag = FALSE;

				/* wait DRQ */	
				while(1) {
					if(StatusPollingCount >= 100000){	/* polling above 1 sec */
						ErrorOccurFlag = TRUE;
						break;
					}

					if (Status & IDE_STATUS_DRQ) {
						break;
					}
					camUStallExecution(1);
					Status = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEStatus);
					StatusPollingCount++;
				}
									
				if(ErrorOccurFlag != TRUE){
					/* get the TAG */
					TAG = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDESectorCount);
					TAG = (TAG >> 3);
					gChnlDrvControlConfig[bChannelNumber].CurrentXferTAG = TAG;
					/* prepare SG */
		
					/* return the correspond packet from gSEQ_INFO */
					ultmpATAPacketAddress=ReturnATA_PacketFromSEQ_INFO(bChannelNumber,SEQID,TAG);
					if (ultmpATAPacketAddress == (PCAM_PACKET)PTR_NULL)
						return;
								
					pATATask=(PATA_TASK)ultmpATAPacketAddress->PacketHead.PTA;
					
					/* write SG table */
					WriteIndexUlong(ATABaseAddr[pATATask->bCh2Chip], iNDEXIDEDestriptorPointer, ultmpATAPacketAddress->ATAPacket.ulPSG);
			
					/* start DMA transfer */
					if (pATATask->ATACmdFlag & camDATA_IN) /* read */
						WriteIndexUchar(ATABaseAddr[pATATask->bCh2Chip], iNDEXIDECICR, (U8)(IDECICR_DMA_Start|SEQID) );		
					else /* write */
						WriteIndexUchar(ATABaseAddr[pATATask->bCh2Chip], iNDEXIDECICR, (U8)(IDECICR_DMA_Start|IDECICR_DMA_RW_Direction|SEQID) );	
				}
			}
			
			if(ErrorOccurFlag == TRUE){
				if (ATA_Timeout_TimerID[bChannelNumber] != ZERO_NULL)
					/* remove ata timeout timer */
					camTimer(Adapter_ID,0,(void *)ATA_Timeout_TimerID[bChannelNumber]);	

				ATA_Timeout(bChannelNumber);
				return;
			}
		}
	}
	else { /* error */

		PCAM_PACKET		ultmpATAPacketAddressHead;
		NeedCheckFIFO[bChannelNumber] = TRUE;
		
		/* soft reset ATA Module after getting status. */
		ATASoftResetModule(bChannelNumber);
		
		/* get the TAG */
		TAG = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDESectorCount);
		TAG = (TAG >> 3);
		gChnlDrvControlConfig[bChannelNumber].UsedTAG &= ~(1 << TAG);
		gChnlDrvControlConfig[bChannelNumber].FreeTAG |= (1 << TAG);
		
		/* get ATA Global control & Status Register */
		ReadPortUlong(ATABaseAddr[bChannelNumber], ATA_STATUS_REGISTER_OFFSET, tmpData);
		
		if (CRC_err)
			ultmpATAPacketAddress=(PCAM_PACKET)PTR_NULL;
		else							
			/* get PCAM_ATA_PACKET from SEQ_INFO */
			ultmpATAPacketAddress=GetATA_PacketFromSEQ_INFO(bChannelNumber,SEQID,TAG);
		
		gChnlDrvControlConfig[bChannelNumber].TCQTransferData=FALSE;
		
		if (ultmpATAPacketAddress != (PCAM_PACKET)PTR_NULL) {

			ultmpATAPacketAddressHead = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;
			ATAModuleBusy[bChannelNumber]=FALSE;
			pCurSEQ_INFO->SEQID2PacketHead=(PVOID)PTR_NULL;
			pCurSEQ_INFO->SEQID2PacketTail=(PVOID)PTR_NULL;
			/* back undo execute packets to wait queue. */
			Back_All_Undone_ATAPackets(bChannelNumber,(PCAM_PACKET)ultmpATAPacketAddressHead);
			ultmpATAPacketAddressHead=(PCAM_PACKET)PTR_NULL;
			gChnlDrvControlConfig[bChannelNumber].FreeTAG = 0xFFFFFFFF;
			gChnlDrvControlConfig[bChannelNumber].UsedTAG = 0x0;
											
			pATATask=(PATA_TASK)ultmpATAPacketAddress->PacketHead.PTA;

			pATATask->ATAStatus = (U16)(tmpData >> 8);
			pATATask->ATAStatus &= ATAERR_HWERROR;
			pATATask->ATAStatus |= ATA_ERROR;						

			GetStatus(ATABaseAddr[pATATask->bCh2Chip], pATATask->Cmd.ATACmd.bStatus);
			
			if (pATATask->Cmd.ATACmd.bStatus & IDE_STATUS_BUSY){
				pATATask->ATAStatus &= ~ATAERR_DRV;
				pATATask->ATAStatus |= ATA_TIMEOUT;
			}
			else{				
				/* set disk status */
				if (pATATask->ATACmdFlag & ATA_DEVICE)									
					ATAErrorStatus(pATATask);
			}		
			
			CAM_wait_until_DMAMachine_Idle(bChannelNumber);
			
			/* free sg resource */
			CAMFreeATASG(bChannelNumber, (PVOID)ultmpATAPacketAddress->PacketHead.PSG);	
			/* free packet resource. */
			CAMFreeCamPacket(bChannelNumber, (PVOID)ultmpATAPacketAddress);
			pATATask->callback(pATATask);	
		}
		else {
			ultmpATAPacketAddressHead = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;
			pCurSEQ_INFO->SEQID2PacketHead=(PVOID)PTR_NULL;
			pCurSEQ_INFO->SEQID2PacketTail=(PVOID)PTR_NULL;			
			/* return all execute commands with error */
			Return_Execute_ATAPackets(bChannelNumber,(PCAM_PACKET)ultmpATAPacketAddressHead,(U16)((tmpData >> 8)|ATA_ERROR));
			ATAModuleBusy[bChannelNumber]=FALSE;
			ultmpATAPacketAddressHead=(PCAM_PACKET)PTR_NULL;	
			gChnlDrvControlConfig[bChannelNumber].FreeTAG = 0xFFFFFFFF;
			gChnlDrvControlConfig[bChannelNumber].UsedTAG = 0x0;						
		}
	}
	
	/* must check module is need to push, By it maybe pushed when no SQEID. */
	for (i=(Adapter_ID*MAX_ATA_CHNL);i<((Adapter_ID+1)*MAX_ATA_CHNL);i++) {
		if (ATAModuleNeedPush[i]==TRUE){
			/* start ata module */
			StartATAOperation(i);
				break;
		}
	}
	return;
}



/*
When the ISR of channel driver is called by OS, it should call CAM_ISR immediately.
Channel driver will tell CAM the interrupt may caused by which adapters. If several adapters 
share the same IRQ, channel driver will set mapped bits of IRQHandle to 1.
CAM will determine if its controllers generated a interrupt. CAM will return FALSE as soon as
possible if this interrupt is not generated by its controllers.
*/
U32 CAM_ISR(U32 IRQHandle)		/* bit#1	The interrupt may caused by adapter# */
{
	U8 Adapter_ID;
	U8 SEQID;
	U8 returnValue=FALSE;	
	U32 INTStatus;
	U8	SEQCntCtrl;
	PSEQ_INFO	pCurSEQ_INFO;
	U8 ModuleType;
	U8	i;
	U8	PlugData;
	U8	MaskFlags;
	PCAM_PACKET ultmpATAPacketAddress;
	PCAM_PACKET ultmpATAPacketAddress_next;
	PATA_TASK pATATask;
	U8 bChannelNumber;
	U8 SEQCntCtrlErr;
	U8 	bdata;	
	U32 tmpData;
				
	/* Flag variable */
	U8 	RestartModule=FALSE;
	U8	FreePacketSource = TRUE;
	U8	RestartFlag=FALSE;
	U8	errorFlag=FALSE;
	U8	SEQerrorFlag=FALSE;
	U8	IntReason=0;
				
	for (Adapter_ID=0;Adapter_ID<MAX_ADAPTER_NUM;Adapter_ID++){
		if (!((IRQHandle >> Adapter_ID)&0x01))
			continue;
		/* channel dirver set the interrupt may caused by adapter# */
		
		/* read Sequence Interrupt Status register, if no sequence INT, return ASAP. */
#ifndef NETWARE
		ReadPortUlong(HostBaseAddr[Adapter_ID], SEQINT_REGISTER_OFFSET, INTStatus);			
#else
		ReadPortUchar(HostBaseAddr[Adapter_ID], SEQINT_REGISTER_OFFSET+1, i);
		INTStatus = (U32)i << 8;
#endif
		if (!(INTStatus & 0xFFFF))		/* check SEQ#0~F INT  */
			continue;	/* no SEQ INT */
		
		/* write 1 clear */
/*
#ifndef NETWARE
		WritePortUlong(HostBaseAddr[Adapter_ID], SEQINT_REGISTER_OFFSET, INTStatus);			
#else
		WritePortUchar(HostBaseAddr[Adapter_ID], SEQINT_REGISTER_OFFSET+1, i);
#endif
*/		
		returnValue = TRUE;

		for (SEQID=START_SEQ_ID;SEQID<MAX_SEQ_ID;SEQID++) {
			if(!INTStatus)
				break;
								
			if ( !(INTStatus & bit(SEQID)) )
				continue;

			INTStatus &= ~bit(SEQID);

#ifndef NETWARE
			WritePortUlong(HostBaseAddr[Adapter_ID], SEQINT_REGISTER_OFFSET, bit(SEQID));			
#else
			WritePortUchar(HostBaseAddr[Adapter_ID], SEQINT_REGISTER_OFFSET+1, (U8)(bit(SEQID-START_SEQ_ID)));			
#endif
			
			RestartModule=FALSE;	/* initial control flag */ 
			SEQerrorFlag = FALSE;	/* initial control flag */
			RestartFlag = FALSE;	/* initial control flag */
				
			/* SEQ# INT */
			pCurSEQ_INFO = &(gSEQ_INFO[Adapter_ID][SEQID]);
			
			/* if is a invalid seqID interrupt. */
			if (pCurSEQ_INFO->SEQID2PacketHead == ((PVOID)PTR_NULL) )
				continue;
						
			ModuleType= (U8)pCurSEQ_INFO->Owner;

			bChannelNumber = Adapter_ID*MAX_ATA_CHNL+ModuleType;			

			if(ATAModuleReceiveTask[bChannelNumber] != TRUE)
				continue;

/***************************************************************************************************/
#ifdef _ERR_INJECT_
	
	if(ErrInj_Buffer_Table[bChannelNumber].count){
		if(ErrInject_SEQ[Adapter_ID] & (1 << SEQID)){		
			if ( ((PPACKET_HEAD)pCurSEQ_INFO->SEQID2PacketHead)->Flag &PACFLAG_NCQ){
				ERR_CAM_NCQ_ISR(Adapter_ID,SEQID);
			}
			/* QUEUED command */
			else if ( ((PPACKET_HEAD)pCurSEQ_INFO->SEQID2PacketHead)->Flag &PACFLAG_QUEUED){
				ERR_CAM_QUEUED_ISR(Adapter_ID,SEQID);
			}			
			else {
				ERR_CAM_NORMAL_ISR(Adapter_ID,SEQID);
			}

			if ( (ATAModuleNeedPush[bChannelNumber]==FALSE) && (ATAModuleBusy[bChannelNumber]==FALSE) ) {
				
				if ((gBOXConfig[bChannelNumber].BOXType != camNOBOX) && (gBOXConfig[bChannelNumber].BOXType & SATAEnclosure)) {
					/* turn off LED */
					CAM_SBOX_HDAccess_LED(bChannelNumber, HDActiveOFF);
				}
			}
			continue;			
		}
	}
#endif
/***************************************************************************************************/

						
			/* NCQ command */
			if ( ((PPACKET_HEAD)pCurSEQ_INFO->SEQID2PacketHead)->Flag &PACFLAG_NCQ){
				CAM_NCQ_ISR(Adapter_ID,SEQID);
			}
			/* QUEUED command */
			else if ( ((PPACKET_HEAD)pCurSEQ_INFO->SEQID2PacketHead)->Flag &PACFLAG_QUEUED){
				CAM_QUEUED_ISR(Adapter_ID,SEQID);
			}			
			else {

				/* remove ata timeout timer */
				if (ATA_Timeout_TimerID[bChannelNumber] != ZERO_NULL) {
					camTimer(Adapter_ID,0,(void *)ATA_Timeout_TimerID[bChannelNumber]);	
					ATA_Timeout_TimerID[bChannelNumber]=ZERO_NULL;
				}
							
				ultmpATAPacketAddress = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;

				/* read Sequence Counter Control Register # */
				ReadPortUchar(HostBaseAddr[Adapter_ID], SEQID*4, SEQCntCtrl);
				SEQCntCtrl &= 0x1F;
							
				/* check error from Sequence Counter Control Register */
				ReadPortUchar(HostBaseAddr[Adapter_ID], (SEQID*4)+2, SEQCntCtrlErr);
				SEQCntCtrlErr &= (1<<ModuleType);
				
				if (SEQCntCtrlErr) {
					SEQerrorFlag = TRUE;
					/* ATAPI PIO SEQCntCtrl will be zero or if error, current packet maybe discount or not */
					if (SEQCntCtrl < pCurSEQ_INFO->PACCNT)
						SEQCntCtrl++;
				}
				else { /* non error */
					if(gAdapterConfig[Adapter_ID].TapeOutVersion < REVISION_2){/* ASIC tapout prior revision 2 */
						if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) {/* SATA */
							#ifdef	_MMIO_
								ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDESATAError, tmpData );
							#else
								ReadIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDESATAError, tmpData );
							#endif
							if (tmpData & CRC_ERR) {	/* CRC error */
								SEQerrorFlag = TRUE;
								pCurSEQ_INFO->PACCNT=SEQCntCtrl;
							}
						}
					}
				}			
								
				if ( ((PPACKET_HEAD)pCurSEQ_INFO->SEQID2PacketHead)->Flag &PACFLAG_ATAPI_PIO){
					if (SEQerrorFlag==TRUE) { /* ATAPI PIO error */
						((PPACKET_HEAD)pCurSEQ_INFO->SEQID2PacketHead)->Flag &= ~PACFLAG_ATAPI_PIO;
						ATAModuleSEQID[bChannelNumber] = U8NULL;
					}
					/* if not error,maybe have next PIO data need to transfer, so not set ATAModuleSEQID[bChannelNumber] = U8NULL*/
				}
				else
					ATAModuleSEQID[bChannelNumber] = U8NULL;

				
				/* if no error(SATA error and packet error) and no data or ide register status need to transfer(PIO, ATA_STATUS_RETURN or ATAPI ), try to execute other packets.*/
				if (!(SEQerrorFlag == TRUE)) {
					
					if (pCurSEQ_INFO->GO==TRUE)
						RestartFlag = TRUE;
					
					if (RestartFlag == TRUE) {
						ATAModuleBusy[bChannelNumber]=FALSE;
						RestartModule = TRUE;
						
						if ( (ATAModuleNeedPush[bChannelNumber]==TRUE) && (ATAModuleBusy[bChannelNumber]==FALSE) ){
							/* start ata module */
							StartATAOperation(bChannelNumber);
						}
					}
				}
				
				ultmpATAPacketAddress = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;
				pCurSEQ_INFO->SEQID2PacketHead=(PVOID)PTR_NULL;
				pCurSEQ_INFO->SEQID2PacketTail=(PVOID)PTR_NULL;
								
				while ((PTR_SIZE)ultmpATAPacketAddress != PTR_NULL){						

					pATATask=(PATA_TASK)ultmpATAPacketAddress->PacketHead.PTA;
					pATATask->ATAStatus= 0;		/* default no error */

					/* initial control flag */
					FreePacketSource = TRUE;
					IntReason = 0;
					errorFlag = FALSE;
							
					if (SEQerrorFlag == TRUE)	
					{
						NeedCheckFIFO[bChannelNumber] = TRUE;
						/* check is error packet */ /* not success done all seq's packets  */
						if (pCurSEQ_INFO->PACCNT == SEQCntCtrl)
						{
							errorFlag = TRUE;								

							/* get ATA Global control & Status Register */
							ReadPortUlong(ATABaseAddr[bChannelNumber], ATA_STATUS_REGISTER_OFFSET, tmpData);
							/* soft reset ATA Module after getting status. */
							ATASoftResetModule(bChannelNumber);
							
							/* if swap box nop command, it will return error at all time, and it will not queue with others packets when execute,
									, needn't abort others packet at waiting queue */
							if (!(pATATask->ATACmdFlag & SWAPBOX_CMD)){
								if (pATATask->ATACmdFlag & ATA_DEVICE) { /* if ATAPI, needn't abort all packets */
									/* back undo packets to wait queue. */
									Back_All_Undone_ATAPackets(bChannelNumber,(PCAM_PACKET)ultmpATAPacketAddress->PacketHead.NPA);
									ultmpATAPacketAddress->PacketHead.NPA=(PVOID)PTR_NULL;
								}
							}
							else /* SWAPBOX_CMD */
								goto SuccessCommandCompleted;
															
							pATATask->ATAStatus = (U16)(tmpData >> 8);

							pATATask->ATAStatus &= ATAERR_HWERROR;
							pATATask->ATAStatus |= ATA_ERROR;						

							/* read 3f6 */
							for ( i = 0; i < 100; i++ ) {
								if (!(ReadIndexUchar(ATABaseAddr[pATATask->bCh2Chip], iNDEXIDEAlternateStatus) & IDE_STATUS_BUSY))
									break; /* wait a while for ATAPI device */
							}
							GetStatus(ATABaseAddr[pATATask->bCh2Chip], pATATask->Cmd.ATACmd.bStatus);
							
							if (pATATask->Cmd.ATACmd.bStatus & IDE_STATUS_BUSY){
								pATATask->ATAStatus &= ~ATAERR_DRV;
								pATATask->ATAStatus |= ATA_TIMEOUT;
							}
							else{				
								/* set disk status */
								if (pATATask->ATACmdFlag & ATA_DEVICE)									
									ATAErrorStatus(pATATask);
								else if (pATATask->ATACmdFlag & SWAPBOX_CMD)
									ATAErrorStatus(pATATask);
								else { 	/* ATAPI device */
									if ( (!(pATATask->ATAStatus &ATAERR_DRV)) && ( pATATask->ATAStatus & ATAERR_OVERRUN ) )
									{	/* that is not a device error */
										pATATask->ATAStatus = 0;
										goto SuccessCommandCompleted;
									}
									if (!(pATATask->ATACmdFlag & INTERNAL_SENSE)){
										/* get ATAPI error , if INTERNAL_SENSE error, ingore it.*/
										ATAPIErrorStatus(pATATask);
										
										/* must have sense buffer so can do internel sense command */
										if ((pATATask->SenseInfoLength > 1)){
											/* If ATAPI packet command error, than send a internel "REQUEST SENSE Command" 
											     to get the sense datas. 	*/
											pATATask->ATACmdFlag = INTERNAL_SENSE;
											IntReason = INT_WAITCMD;
											goto SuccessCommandCompleted;
										}
									}
									else {
										pATATask->ATACmdFlag &= ~INTERNAL_SENSE;
										ATAPIErrorStatus(pATATask); /* even request sense is failed. */
 										ReadIndexUchar(ATABaseAddr[pATATask->bCh2Chip], iNDEXIDEStatus); /*clear possible device INT */
 									}											
								}
							}
							goto FailCommandCompleted;
						}
						else
						{
							pCurSEQ_INFO->PACCNT--;
						}
					}
					else { /* success done all seq's packets */
						pATATask->ATAStatus= 0;
						
						/* get INT reason */
						if (!(pATATask->ATACmdFlag & ATA_DEVICE))
							IntReason = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDESectorCount)& 0x07;
							
						/* If the packet's ATACmdFlag be set with ATA_STATUS_RETURN , it only can queued at the tail of execute packets queue(or only itself at the queue),
		     				   so when error or success completed, will return ata registers statses.*/
						if (pATATask->ATACmdFlag & ATA_STATUS_RETURN) /* need read it's ata registers statses when command success completed */
							ATAErrorStatus(pATATask);
					}
SuccessCommandCompleted:					
				
					
					if(IntReason == INT_WAITCMD){
						FreePacketSource = FALSE;
						SubmitInternelSense(pATATask,ultmpATAPacketAddress);
					}
					else if ( pATATask->ATACmdFlag & PIO_XFER )
					{
						if (!(pATATask->ATACmdFlag & ATA_DEVICE) ) { /* ATAPI */
							U32	DataTransferLengthTotal=0;
							U32	DataTransferLengthTotalTemp=0;
							U8 PIOStatus=camSUCCESS;
							
							bdata = ReadIndexUchar(ATABaseAddr[pATATask->bCh2Chip], iNDEXIDEStatus); /* have data to be trasfer */ /*clear possible device INT */
							if (bdata & IDE_STATUS_DRQ){
								
								if ( gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA ) {
									ReadIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_PIO_Transfer_count, DataTransferLengthTotalTemp);
									DataTransferLengthTotalTemp &= 0x0000FFFF;									
								}
								else {
									DataTransferLengthTotalTemp = (U32)ReadIndexUchar(ATABaseAddr[pATATask->bCh2Chip], iNDEXIDECylinderLow)
							                   | ((U32)ReadIndexUchar(ATABaseAddr[pATATask->bCh2Chip], iNDEXIDECylinderHigh) << 8);
								}
								DataTransferLengthTotalTemp >>= 1;  
								DataTransferLengthTotal = DataTransferLengthTotal+DataTransferLengthTotalTemp;

								if (DataTransferLengthTotalTemp==0) { /* no datas to be transfer */
									/* soft reset ATA Module after getting status. */
									ATASoftResetModule(bChannelNumber);
									pATATask->ATAStatus = ATA_ERROR;
									errorFlag=TRUE;
									((PPACKET_HEAD)ultmpATAPacketAddress)->Flag &= ~PACFLAG_ATAPI_PIO;
									ATAModuleSEQID[bChannelNumber] = U8NULL;									
									goto FailCommandCompleted;
								}

								if ( pATATask->ATACmdFlag & camDATA_IN ){
									
  									PIOStatus = ReadBuffer(pATATask->bCh2Chip,(PU16)pATATask->DataBuffer, DataTransferLengthTotalTemp );
								}
								else {
									PIOStatus = WriteBuffer(pATATask->bCh2Chip,(PU16)pATATask->DataBuffer, DataTransferLengthTotalTemp );
    								}
								
								/* check error from Sequence Counter Control Register */
								ReadPortUchar(HostBaseAddr[Adapter_ID], (SEQID*4)+2, SEQCntCtrlErr);
								SEQCntCtrlErr &= (1<<ModuleType);
								
								/* after PIO data transfer, maybe be with error */
								if (SEQCntCtrlErr || (PIOStatus == camFAIL)) {
									/* soft reset ATA Module after getting status. */
									ATASoftResetModule(bChannelNumber);
									pATATask->ATAStatus = ATA_ERROR;
									errorFlag=TRUE;
									((PPACKET_HEAD)ultmpATAPacketAddress)->Flag &= ~PACFLAG_ATAPI_PIO;
									ATAModuleSEQID[bChannelNumber] = U8NULL;									
									goto FailCommandCompleted;
								}
													
								(PU16)pATATask->DataBuffer = (PU16)pATATask->DataBuffer+DataTransferLengthTotal;

								/* need next ISR to clear ATAPI interrupt */
								pCurSEQ_INFO->SEQID2PacketHead=(PVOID)ultmpATAPacketAddress;
								pCurSEQ_INFO->SEQID2PacketTail=(PVOID)ultmpATAPacketAddress;									
								goto NextSEQID;
								/*((PPACKET_HEAD)ultmpATAPacketAddress)->Flag &= ~PACFLAG_ATAPI_PIO;*/
								/*ATAModuleSEQID[bChannelNumber] = U8NULL;*/
							}
							else{/* else of if (bdata & IDE_STATUS_DRQ) */
								if ( pATATask->ATACmdFlag & INTERNAL_SENSE )
									pATATask->ATAStatus = ATA_ERROR + ATAERR_DRV;
								pATATask->ATACmdFlag &= ~INTERNAL_SENSE;
								((PPACKET_HEAD)ultmpATAPacketAddress)->Flag &= ~PACFLAG_ATAPI_PIO;
								ATAModuleSEQID[bChannelNumber] = U8NULL;	
							}
						}
						else { /* ATA */
							U32	DataTransferLength = 0;
							U8 PIOStatus;
							
							DataTransferLength = (pATATask->DataTransferLength >> 1);
							if ( pATATask->ATACmdFlag & camDATA_IN ){
  								PIOStatus = ReadBuffer(pATATask->bCh2Chip,(PU16)pATATask->DataBuffer, DataTransferLength );
							}
							else {
								PIOStatus = WriteBuffer(pATATask->bCh2Chip,(PU16)pATATask->DataBuffer, DataTransferLength );
    							}		
						
							if(PIOStatus == camFAIL)
								pATATask->ATAStatus = ATA_ERROR;
						}
					}	/*end of (pATATask->ATACmdFlag & PIO_XFER )*/									
FailCommandCompleted:				
					if (FreePacketSource==TRUE) {
						/* free sg resource */
						CAMFreeATASG(bChannelNumber, (PVOID)ultmpATAPacketAddress->PacketHead.PSG);	
							/* free packet resource. */
						CAMFreeCamPacket(bChannelNumber, (PVOID)ultmpATAPacketAddress);
					}

					if ( !(pATATask->ATACmdFlag & INTERNAL_SENSE) ) {
						pATATask->callback(pATATask);	
					}	
											
					ultmpATAPacketAddress_next = (PCAM_PACKET)ultmpATAPacketAddress->PacketHead.NPA;
					
					if (errorFlag==TRUE)
						break;
						
					/* next packet */
					ultmpATAPacketAddress = ultmpATAPacketAddress_next;
					
				}/* end of while ((U32)ultmpATAPacketAddress != PTR_NULL)*/

				if (RestartModule==FALSE) 
					ATAModuleBusy[bChannelNumber]=FALSE;

				CAMFreeSEQID(Adapter_ID,SEQID);	

				/* must check module is need to push, By it maybe pushed when no SQEID. */
				for (i=(Adapter_ID*MAX_ATA_CHNL);i<((Adapter_ID+1)*MAX_ATA_CHNL);i++) {
					if ( (ATAModuleNeedPush[i]==TRUE) && (ATAModuleBusy[i]==FALSE) ){
						/* start ata module */
						StartATAOperation(i);
						break;
					}
				}
			} /* end of NCQ/QUEUED command */	

			if ( (ATAModuleNeedPush[bChannelNumber]==FALSE) && (ATAModuleBusy[bChannelNumber]==FALSE) ) {
				
				if ((gBOXConfig[bChannelNumber].BOXType != camNOBOX) && (gBOXConfig[bChannelNumber].BOXType & SATAEnclosure)) {
					/* turn off LED */
					CAM_SBOX_HDAccess_LED(bChannelNumber, HDActiveOFF);
				}
			}
NextSEQID:	;										
		}/* end of for (SEQID=START_SEQ_ID;SEQID<MAX_SEQ_ID;SEQID++) */
	}/* end of Adapter_ID */

	if (returnValue == TRUE)
		return TRUE;

	/* check is I2C interrupt */
	for (Adapter_ID=0;Adapter_ID<MAX_ADAPTER_NUM;Adapter_ID++) {
		U16 wdata;
		
		if (!((IRQHandle >> Adapter_ID)&0x01) )
			continue;
		
		if (I2CBusy[Adapter_ID] != TRUE)
			continue;
			
		ReadPortWord(HostBaseAddr[Adapter_ID], oFFSETSBusControlStatus, wdata );
		if (wdata & 0x0006)	/* ERR */
		{
			CAM_ResetSerialBusMachine(Adapter_ID);
		}
				
		if (wdata&0x0F00)
			continue;		
		if (!(wdata&0x000F))
			continue;		
			
		/* mask ERR & Finish and clear ERR & Finish */
		WritePortWord(HostBaseAddr[Adapter_ID], oFFSETSBusControlStatus, 0x0F0F);

		I2CQueueCurrent[Adapter_ID].callback(&I2CQueueCurrent[Adapter_ID]);

		I2CBusy[Adapter_ID] =FALSE;		
		CAM_StartI2C(Adapter_ID);
		return TRUE;		
	}
	
	for (Adapter_ID=0;Adapter_ID<MAX_ADAPTER_NUM;Adapter_ID++) {
		if (!((IRQHandle >> Adapter_ID)&0x01) )
			continue;
		
		/* check unplug flag */
		ReadPortUchar(HostBaseAddr[Adapter_ID], oFFSETHOSTSATAPLUG+2, MaskFlags);
		MaskFlags &= 0xF0;
		for ( i=0; i < MAX_ATA_CHNL; i++){
			if ( gChnlConfig[Adapter_ID*MAX_ATA_CHNL+i].ChnlType == CHNL_SATA)
			{
				/* read SATA plug control and status register */
				ReadPortUchar(HostBaseAddr[Adapter_ID], oFFSETHOSTSATAPLUG, PlugData);
				PlugData &= (U8)1 << i;
				PlugData &= ~MaskFlags;
				if (!PlugData)
					continue;
				/* clear unplug flags */
				WritePortUchar(HostBaseAddr[Adapter_ID], oFFSETHOSTSATAPLUG, PlugData);
				returnValue = TRUE;

				/* check if the SATA cable is re-connected again */
				if ( CAM_HardResetSATAChannel((U8)(Adapter_ID*MAX_ATA_CHNL+i)) == camSUCCESS ) {
					CAM_Clear_UnmaskPlugUnplugFlags((U8)(Adapter_ID*MAX_ATA_CHNL+i));
					if ( gDrvConfig[Adapter_ID*MAX_ATA_CHNL+i][0].DevFlag == DEV_NOTFOUND) {
						if((gBOXConfig[(Adapter_ID*MAX_ATA_CHNL+i)].BOXType != camNOBOX) && (gBOXConfig[Adapter_ID*MAX_ATA_CHNL+i].BOXType & (PSS10398_1_CHNL|SuperSwap1100)))
							PlugOff_Flag[(Adapter_ID*MAX_ATA_CHNL+i)] = FALSE;						
						cam_SetStatusByPlug_unplug((U8) (Adapter_ID*MAX_ATA_CHNL+i), 0, NEW_DRIVE);
					}
					continue;
				}
				CAM_Clear_UnmaskPlugUnplugFlags((U8)(Adapter_ID*MAX_ATA_CHNL+i));
				
				/* turn off the LED of Enclosure */
				if ((gBOXConfig[Adapter_ID*MAX_ATA_CHNL+i].BOXType != camNOBOX) && (gBOXConfig[Adapter_ID*MAX_ATA_CHNL+i].BOXType & SATAEnclosure)) {
					CAM_SBOX_HDAccess_LED((U8) (Adapter_ID*MAX_ATA_CHNL+i), HDActiveOFF);					
					if((gBOXConfig[Adapter_ID*MAX_ATA_CHNL+i].BOXType & (PSS10398_1_CHNL|SuperSwap1100))){
						gBOXStatus[(Adapter_ID*MAX_ATA_CHNL+i)].PowerState = PowerOFF;
						PlugOff_Flag[(Adapter_ID*MAX_ATA_CHNL+i)] = TRUE;
					}
				}
				CAM_Clear_UnmaskPlugUnplugFlags((U8)(Adapter_ID*MAX_ATA_CHNL+i));
								
				if ( gDrvConfig[Adapter_ID*MAX_ATA_CHNL+i][0].DevFlag != DEV_NOTFOUND) {
					gDrvConfig[Adapter_ID*MAX_ATA_CHNL+i][0].DevFlag = DEV_NOTFOUND;
					cam_SetStatusByPlug_unplug((U8) (Adapter_ID*MAX_ATA_CHNL+i), 0, NO_DRIVE);
				}
			}
		}
				
		/* check plug flag */
		ReadPortUchar(HostBaseAddr[Adapter_ID], oFFSETHOSTSATAPLUG+2, MaskFlags);
		MaskFlags &= 0x0F;
		for ( i=0; i < MAX_ATA_CHNL; i++){
			if ( gChnlConfig[Adapter_ID*MAX_ATA_CHNL+i].ChnlType == CHNL_SATA)
			{
				/* read SATA plug control and status register */
				ReadPortUchar(HostBaseAddr[Adapter_ID], oFFSETHOSTSATAPLUG, PlugData);
				PlugData &= (0x10 << i);
				PlugData &= ~MaskFlags;
				if (!PlugData)
					continue;
				/* clear plug flags */
				WritePortUchar(HostBaseAddr[Adapter_ID], oFFSETHOSTSATAPLUG, PlugData);
				returnValue = TRUE;
				
				if ( CAM_HardResetSATAChannel((U8)(Adapter_ID*MAX_ATA_CHNL+i)) == camDEVICE_NOT_FOUND ) {
					CAM_Clear_UnmaskPlugUnplugFlags((U8)(Adapter_ID*MAX_ATA_CHNL+i));
					gDrvConfig[Adapter_ID*MAX_ATA_CHNL+i][0].DevFlag = DEV_NOTFOUND;
					/* turn off the LED of Enclosure */
					if ((gBOXConfig[Adapter_ID*MAX_ATA_CHNL+i].BOXType != camNOBOX) && (gBOXConfig[Adapter_ID*MAX_ATA_CHNL+i].BOXType & SATAEnclosure)) {
						CAM_SBOX_HDAccess_LED((U8) (Adapter_ID*MAX_ATA_CHNL+i), HDActiveOFF);					
						if((gBOXConfig[Adapter_ID*MAX_ATA_CHNL+i].BOXType & (PSS10398_1_CHNL|SuperSwap1100))){
							gBOXStatus[(Adapter_ID*MAX_ATA_CHNL+i)].PowerState = PowerOFF;
							PlugOff_Flag[(Adapter_ID*MAX_ATA_CHNL+i)] = TRUE;
						}
					}					
					continue;
				}
				CAM_Clear_UnmaskPlugUnplugFlags((U8)(Adapter_ID*MAX_ATA_CHNL+i));				

				if((gBOXConfig[(Adapter_ID*MAX_ATA_CHNL+i)].BOXType != camNOBOX) && (gBOXConfig[Adapter_ID*MAX_ATA_CHNL+i].BOXType & (PSS10398_1_CHNL|SuperSwap1100)))
					PlugOff_Flag[(Adapter_ID*MAX_ATA_CHNL+i)] = FALSE;
										
				if ( gDrvConfig[Adapter_ID*MAX_ATA_CHNL+i][0].DevFlag == DEV_NOTFOUND) 
					cam_SetStatusByPlug_unplug((U8) (Adapter_ID*MAX_ATA_CHNL+i), 0, NEW_DRIVE);
			}
		}
		
	}

	return returnValue;
}
